/**
 * ==============================================
 *  DQL : Descartes Query Language
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Juergen Walter, Fabian Gorsler, Fabian Brosig and other contributors.
 *
 * Project Info:   http://descartes.tools/dql
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.connector.nop;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Properties;

import org.eclipse.emf.ecore.util.EcoreUtil;

import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public class PerformanceMetricsQueryConnectorImpl extends AbstractNopQueryConnector
		implements PerformanceMetricsQueryConnector {
	ArrayList<String> constraintNames = new ArrayList<String>();
	ArrayList<String> doFConstraintNames = new ArrayList<String>();
	ArrayList<String> explorationStrategyNames = new ArrayList<String>();
	private String constraint = null;
	private String doFConstraint = null;
	private String exploratationStrategy = null;
	private Properties explorationStrategyProperties;
	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;

	public PerformanceMetricsQueryConnectorImpl() {
		super();
		loadConstraintNames();
		loadDoFConstraintNames();
		loadExplorationStrategyNames();
	}

	@Override
	public String getConstraint() {
		return constraint;
	}

	@Override
	public String getDoFConstraint() {
		return doFConstraint;
	}

	@Override
	public String getDoFExplorationStrategy() {
		return exploratationStrategy;
	}

	@Override
	public Properties getDoFExplorationStrategyProperties() {
		return explorationStrategyProperties;
	}

	@Override
	public List<String> getSupportedConstraintNames() {
		return constraintNames;
	}

	@Override
	public List<String> getSupportedDoFConstraintNames() {
		return doFConstraintNames;
	}

	@Override
	public List<String> getSupportedExplorationStrategyNames() {
		return explorationStrategyNames;
	}

	@Override
	public List<String> getSupportedTimeUnits() {
		// TODO Auto-generated method stub
		return null;
	}

	private void loadConstraintNames() {
		constraintNames.add("DEFAULT");
		constraintNames.add("FAST");
		constraintNames.add("ACCURATE");
		constraintNames.add("TIME_1s");
		constraintNames.add("TIME_5s");
		constraintNames.add("ACCURACY_95");
		constraintNames.add("ACCURACY_99");
	}

	private void loadDoFConstraintNames() {
		doFConstraintNames.add("DEFAULT");
		doFConstraintNames.add("FAST");
		doFConstraintNames.add("ACCURATE");
		doFConstraintNames.add("TIME_1s");
		doFConstraintNames.add("TIME_5s");
		doFConstraintNames.add("ACCURACY_95");
		doFConstraintNames.add("ACCURACY_99");
	}

	private void loadExplorationStrategyNames() {
		explorationStrategyNames.add("FULL");
		explorationStrategyNames.add("RANDOM");
	}

	@Override
	public List<EntityMapping> requestDoFMetrics(final EntityMapping request) {
		fakeProgress();
		final ArrayList<EntityMapping> results = new ArrayList<EntityMapping>();

		for (@SuppressWarnings("unused")
		final DoF doF : request.getDoF()) {
			results.add(requestMetrics(request));
		}

		return results;
	}

	@Override
	public EntityMapping requestMetrics(final EntityMapping request) {
		fakeProgressIfNotDoF();
		final EntityMapping response = EcoreUtil.copy(request);
		final ArrayList<Entity> entities = new ArrayList<Entity>();
		entities.addAll(response.getResources());
		entities.addAll(response.getServices());

		for (final Entity entity : entities) {
			final ArrayList<DecimalResult> results = new ArrayList<DecimalResult>();

			for (final Probe probe : entity.getProbes()) {
				final DecimalResult result = domainFactory.createDecimalResult();

				result.setMetricName(probe.getMetricName());
				result.setValid(true);
				result.setAccuracy(new BigDecimal(0.9995f));
				result.setValue(new BigDecimal(1.234f));

				results.add(result);
			}

			entity.getProbes().clear();
			entity.getProbes().addAll(results);
		}

		return response;
	}

	@Override
	public void setConstraint(final String constraint) {
		// Although being not relevant for results, the value is stored and
		// returned by the getter, e.g. for regression testing.
		if (constraintNames.contains(constraint)) {
			this.constraint = constraint;
		} else {
			this.constraint = null;
		}
	}

	@Override
	public void setDoFConstraint(final String doFConstraint) {
		if (doFConstraintNames.contains(doFConstraint)) {
			this.doFConstraint = doFConstraint;
		} else {
			this.doFConstraint = null;
		}
	}

	@Override
	public void setDoFExplorationStrategy(final String strategy, final Properties properties) {
		if (explorationStrategyNames.contains(strategy)) {
			exploratationStrategy = strategy;
		} else {
			exploratationStrategy = null;
		}

		explorationStrategyProperties = properties;
	}

	@Override
	public void setObservationPeriod(final Date start, final Date end, final long sampleSize) {
		// TODO Auto-generated method stub

	}

	@Override
	public void setObservationPeriod(final long amount, final String unit, final long sampleSize) {
		// TODO Auto-generated method stub

	}

	@Override
	public void setObservationReference(final String type, final String identifier, final long sampleSize) {

	}

	@Override
	public boolean supportsConstraintRequests() {
		return true;
	}

	@Override
	public boolean supportsDoFRequests() {
		return true;
	}

	@Override
	public boolean supportsMetricRequests() {
		return true;
	}

	@Override
	public boolean supportsObservationPeriods() {
		return false;
	}

	@Override
	public void reset() {
		// Nothing to do here
	}
	
	private void fakeProgressIfNotDoF() {
		StackTraceElement[] stackTrace = Thread.currentThread().getStackTrace();
		boolean isCalledByDoF = false;
		for (int i = 0; i < stackTrace.length; i++) {
			if (stackTrace[i].getMethodName().equals("requestDoFMetrics")) {
				isCalledByDoF = true;
				break;
			}
		}

		if (!isCalledByDoF) {
			fakeProgress();
		}
	}
	
}
