/**
 * ==============================================
 *  DQL : Descartes Query Language
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Juergen Walter, Fabian Gorsler, Fabian Brosig and other contributors.
 *
 * Project Info:   http://descartes.tools/dql
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.core.engine.aggregates;

import java.math.BigDecimal;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;

import tools.descartes.dql.core.engine.aggregates.calculators.AggregateCalculator;
import tools.descartes.dql.core.engine.aggregates.calculators.AggregateWithProperties;
import tools.descartes.dql.core.engine.util.AggregateHelper;
import tools.descartes.dql.core.engine.util.DQLLogger;
import tools.descartes.dql.core.engine.util.QueryParserHelper;
import tools.descartes.dql.lang.descartesQL.AggregateMetricClause;
import tools.descartes.dql.lang.descartesQL.ConfigurationPropertiesClause;
import tools.descartes.dql.lang.descartesQL.MetricClause;
import tools.descartes.dql.lang.descartesQL.MetricClauses;
import tools.descartes.dql.lang.descartesQL.MetricStarClause;
import tools.descartes.dql.lang.descartesQL.StatisticalAggregateType;
import tools.descartes.dql.models.mapping.domain.Aggregate;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

/**
 * Implementation of AggregateProcessor
 * 
 * @author Fabian Gorsler <fabian.gorsler@student.kit.edu>
 */
public class AggregateProcessorImpl implements AggregateProcessor {
	private final Logger log = DQLLogger.getLogger(this.getClass().getName());
	private AggregateMetricClause clause = null;
	private String clauseString = null;
	private Map<String, Set<String>> requiredProbes;
	private StatisticalAggregateType function;
	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;
	private Properties properties;

	public AggregateProcessorImpl(final AggregateMetricClause clause) {
		this.clause = clause;
	}

	private Map<String, Set<String>> interpretMetricClause(final EObject metrics, final Map<String, Entity> aliasMap) {
		if (metrics instanceof MetricStarClause) {
			if (log.isDebugEnabled()) {
				log.debug("Interpreting a MetricStarClause: " + metrics);
			}

			return interpretMetrics((MetricStarClause) metrics, aliasMap);
		} else if (metrics instanceof MetricClauses) {
			if (log.isDebugEnabled()) {
				log.debug("Interpreting a MetricReferencesClause: " + metrics);
			}

			return interpretMetrics((MetricClauses) metrics, aliasMap);
		}

		log.error("Unknown metric clause type detected: " + metrics);
		return null;
	}

	private Map<String, Set<String>> interpretMetrics(final MetricClauses metrics, final Map<String, Entity> aliasMap) {
		final EList<MetricClause> clauses = metrics.getMetrics();
		final Map<String, Set<String>> requiredMetrics = new HashMap<String, Set<String>>(clauses.size());

		for (final MetricClause m : clauses) {
			final String entityRef = m.getMetricReference().getEntityReference();
			final String metricRef = m.getMetricReference().getMetricReference();
			final Entity entity = aliasMap.get(entityRef);
			final String identifier;

			// Resolve identifier of metric
			if (entity != null) {
				identifier = entity.getIdentifier();
			} else {
				identifier = entityRef;
			}

			AggregateHelper.createOrUpdate(requiredMetrics, identifier, metricRef);
		}

		return requiredMetrics;
	}

	private Map<String, Set<String>> interpretMetrics(final MetricStarClause metrics, final Map<String, Entity> aliasMap) {
		final String metricName = metrics.getMetricName();
		final Map<String, Set<String>> requiredMetrics = new HashMap<String, Set<String>>(aliasMap.size());

		// Add all aliased entities to the list of required metrics
		for (final Entity e : aliasMap.values()) {
			final Set<String> m = new HashSet<String>(1);

			m.add(metricName);

			requiredMetrics.put(e.getIdentifier(), m);
		}

		return requiredMetrics;
	}

	@Override
	public void postProcess(final EntityMapping mapping, final Map<String, Entity> entityIndex,
			final Map<String, Map<String, Probe>> probeIndex) {
		final AggregateCalculator calc = AggregateCalculatorFactory.createCalculator(function);
		final EList<Aggregate> aggregates = mapping.getAggregates();
		final Aggregate aggregate = domainFactory.createAggregate();
		final DecimalResult aggregateResult = domainFactory.createDecimalResult();
		boolean isValid = true;

		if (calc instanceof AggregateWithProperties) {
			if (properties == null) {
				log.error("Aggregate requires additional properties: " + function);
				return;
			}

			((AggregateWithProperties) calc).setProperties(properties);
		}

		isValid = processEntities(probeIndex, calc, isValid);

		aggregates.add(aggregate);

		aggregate.setAlias(clauseString);
		aggregate.getProbes().add(aggregateResult);

		aggregateResult.setAccuracy(new BigDecimal(1));
		aggregateResult.setMetricName(function.getLiteral());
		aggregateResult.setValid(isValid);
		aggregateResult.setValue(calc.getResult());
	}

	@Override
	public Map<String, Set<String>> preProcess(final EntityMapping mapping, final Map<String, Entity> aliasMap) {
		final EObject metrics = clause.getMetrics();
		requiredProbes = interpretMetricClause(metrics, aliasMap);
		function = clause.getFunction();
		final ConfigurationPropertiesClause configurationProperties = clause.getProperties();

		if (configurationProperties != null) {
			properties = QueryParserHelper.toFlatConfigurationPropertiesClause(configurationProperties);
		}

		// Not longer required. Allow GC to clean up.
		clauseString = clause.toString();
		clause = null;

		if (requiredProbes == null) {
			log.error("Interpreting required metrics was not successful. Aborting.");
			return null;
		}

		return requiredProbes;
	}

	private boolean processEntities(final Map<String, Map<String, Probe>> probeIndex, final AggregateCalculator calc,
			boolean isValid) {
		for (final String identifier : requiredProbes.keySet()) {
			final Map<String, Probe> probesOfEntity = probeIndex.get(identifier);
			isValid = processProbes(calc, isValid, identifier, probesOfEntity);
		}

		return isValid;
	}

	private boolean processProbes(final AggregateCalculator calc, boolean isValid, final String identifier,
			final Map<String, Probe> probesOfEntity) {
		for (final String metricName : requiredProbes.get(identifier)) {
			final Probe probe = probesOfEntity.get(metricName);
			final DecimalResult result;

			if (probe == null || !(probe instanceof DecimalResult)) {
				log.warn("Calculating aggregate yields errors, probe causing errors: " + probe);
				isValid = false;
				continue;
			}

			result = (DecimalResult) probe;

			if (!result.getValid()) {
				log.warn("Result value was marked as being invalid for result: " + result);
				isValid = false;
			}

			calc.addValue(result.getValue());
		}

		// Report if the EntityMapping and the calculation process are valid
		return isValid && calc.isValid();
	}

}
