/**
 * ==============================================
 *  DQL : Descartes Query Language
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Juergen Walter, Fabian Gorsler, Fabian Brosig and other contributors.
 *
 * Project Info:   http://descartes.tools/dql
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.core.engine.observations;

import java.util.Date;

import org.apache.log4j.Logger;

import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.core.engine.util.DQLLogger;
import tools.descartes.dql.lang.descartesQL.AbsoluteTimeClause;
import tools.descartes.dql.lang.descartesQL.BetweenTimeClause;
import tools.descartes.dql.lang.descartesQL.ObserveBetweenClause;
import tools.descartes.dql.lang.descartesQL.RelativeTimeClause;

public class ObserveBetweenInterpreter implements ObservationInterpreter {
	private final Logger log = DQLLogger.getLogger(this.getClass().getName());
	private ObserveBetweenClause observation = null;
	private ObservationSpecification observationSpecification = null;

	public ObserveBetweenInterpreter(final ObserveBetweenClause observation) {
		this.observation = observation;
		interpretDate();
	}

	private TimeCalculator calcLeftAbsoluteRightAbsolute(final BetweenTimeClause endingTime) {
		// Right value can be calculated independently
		TimeCalculator endingTimeCalc;
		endingTimeCalc = new AbsoluteTimeCalculator((AbsoluteTimeClause) endingTime);
		return endingTimeCalc;
	}

	private TimeCalculator calcLeftAbsoluteRightRelative(final BetweenTimeClause endingTime,
			final TimeCalculator startingTimeCalc) throws Exception {
		// Right value can be calculated independently
		TimeCalculator endingTimeCalc;
		if (!startingTimeCalc.isValid()) {
			// Without a valid starting time, the ending time cannot be
			// calculated
			throw new Exception();
		}

		endingTimeCalc = new RelativeTimeCalculator();
		endingTimeCalc.setReferenceDate(startingTimeCalc.getDate());
		endingTimeCalc.parseDate((RelativeTimeClause) endingTime);
		return endingTimeCalc;
	}

	private TimeCalculator calcLeftRelativeRightAbsolute(final BetweenTimeClause startingTime,
			final BetweenTimeClause endingTime, final TimeCalculator startingTimeCalc) throws Exception {
		TimeCalculator endingTimeCalc;
		endingTimeCalc = calcLeftAbsoluteRightAbsolute(endingTime);

		if (!endingTimeCalc.isValid()) {
			throw new Exception();
		}

		startingTimeCalc.setReferenceDate(endingTimeCalc.getDate());
		startingTimeCalc.setBackwards(true);
		startingTimeCalc.parseDate((RelativeTimeClause) startingTime);
		return endingTimeCalc;
	}

	private TimeCalculator calcLeftRelativeRightRelative(final BetweenTimeClause startingTime,
			final BetweenTimeClause endingTime, final TimeCalculator startingTimeCalc) throws Exception {
		TimeCalculator endingTimeCalc;
		// Right value relative -> current time as reference for
		// negative relative starting time
		final Date currentTime = new Date();
		startingTimeCalc.setReferenceDate(currentTime);
		startingTimeCalc.setBackwards(true);
		startingTimeCalc.parseDate((RelativeTimeClause) startingTime);

		endingTimeCalc = calcLeftAbsoluteRightRelative(endingTime, startingTimeCalc);
		return endingTimeCalc;
	}

	@Override
	public void configureConnector(final PerformanceMetricsQueryConnector queryConnector, final long sampleSize) {
		if (observationSpecification == null) {
			log.error("No observation specification set.");
			return;
		}

	}

	@Override
	public void interpretDate() {
		final BetweenTimeClause startingTime = observation.getStartingTime();
		final BetweenTimeClause endingTime = observation.getEndingTime();
		TimeCalculator startingTimeCalc = null;
		TimeCalculator endingTimeCalc = null;
		boolean leftAbsolute = false;
		boolean rightAbsolute = false;

		if (startingTime instanceof AbsoluteTimeClause) {
			leftAbsolute = true;
		}

		if (endingTime instanceof AbsoluteTimeClause) {
			rightAbsolute = true;
		}

		try {

			if (leftAbsolute) {
				// left is absolute
				startingTimeCalc = calcLeftAbsoluteRightAbsolute(startingTime);

				if (rightAbsolute) {
					endingTimeCalc = calcLeftAbsoluteRightAbsolute(endingTime);
				} else {
					endingTimeCalc = calcLeftAbsoluteRightRelative(endingTime, startingTimeCalc);
				}

			} else {
				// left is relative
				startingTimeCalc = new RelativeTimeCalculator();

				if (rightAbsolute) {
					endingTimeCalc = calcLeftRelativeRightAbsolute(startingTime, endingTime, startingTimeCalc);
				} else {
					endingTimeCalc = calcLeftRelativeRightRelative(startingTime, endingTime, startingTimeCalc);
				}

			}

			if (startingTimeCalc.isValid() && endingTimeCalc.isValid()) {
				storeResult(startingTimeCalc, endingTimeCalc);
			}

		} catch (final Exception e) {
			return;
		}
	}

	private void storeResult(final TimeCalculator startingTimeCalc, final TimeCalculator endingTimeCalc) {
		final Date startDate = startingTimeCalc.getDate();
		final Date endDate = endingTimeCalc.getDate();

		observationSpecification = new ObservationSpecification(startDate, endDate);
	}

}
