package tools.descartes.dql.core.engine.whatif;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;

import org.eclipse.emf.ecore.util.EcoreUtil;

import tools.descartes.dql.connector.ModelStructureQueryConnector;
import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;

import tools.descartes.dql.core.engine.interpreter.QueryInterpreter;
import tools.descartes.dql.lang.descartesQL.WhatIfQuery;
import tools.descartes.dql.lang.descartesQL.WhatIfVariationClause;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;
import tools.descartes.dql.models.mapping.mapping.MappingFactory;

public class WhatIfQueryInterpreter implements QueryInterpreter {

	private final MappingFactory mappingFactory = MappingFactory.eINSTANCE;
	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;

	private List<EntityMapping> results = new LinkedList<EntityMapping>();

	private ModelStructureQueryConnector modelStructureQuerConnector = null;
	private PerformanceMetricsQueryConnector performanceMetricsConnector = null;

	// Connector State
	private WhatIfQuery query;
	private String modelLocation;
	private List<QueryProgressListener> listeners = new ArrayList<QueryProgressListener>();

	// constructor
	public WhatIfQueryInterpreter(final WhatIfQuery query) {
		this.query = query;
		this.modelLocation = query.getUsingClause().getModelReference().getModel().getLocation();
	}

	// constructor
	public WhatIfQueryInterpreter(final WhatIfQuery query, final List<QueryProgressListener> listeners) {
		this.query = query;
		this.listeners = listeners;
		this.modelLocation = query.getUsingClause().getModelReference().getModel().getLocation();
	}

	@Override
	public EntityMapping getResult() {
		return null;
		// there is not just one result, because of multiple results
	}

	@Override
	public List<EntityMapping> getResults() {
		return results;
	}

	@Override
	public boolean hasMultipleResults() {
		return true;
	}

	@Override
	public boolean interpretQuery() {

		/*
		 * no need to split interpretation into sub-mehtods like
		 * PerformanceMetricsQueryInterpreter.
		 * 
		 * QueryExecutionEngineImpl checks if the query is a WhatIfQuery or not
		 * => no need to check again. Called methods are checking step by step
		 * if query input is valid or not.
		 * 
		 */

		initializeConnectors(query.getUsingClause().getModelReference().getFamily().getFamily());

		validateDoFs();

		EntityMapping requestWithoutDoF = createRequestWithoutDof();
		EntityMapping requestWithDoF = addDoFToRequest(requestWithoutDoF);

		// intercept results for filtering and customize them
		this.results = performanceMetricsConnector.requestDoFMetrics(requestWithDoF);

		if (query.getFilteringClause() != null)
			filterResults();

		modifyToDeltas();

		return true;
	}

	/*
	 * modifies results that results.get(i) for i > 0 has delta-values as result
	 */
	private void modifyToDeltas() {
		// TODO Auto-generated method stub

	}

	/*
	 * processes the filtering
	 */
	private void filterResults() {
		WhatIfFilter filter = WhatIfFilterFactory.parseFilter(query.getFilteringClause());
		EntityMapping initial = results.get(0);
		List<EntityMapping> filteredResults = new LinkedList<EntityMapping>();
		for (EntityMapping result : results)
			filteredResults.add(filter.applyFilter(initial, result));

	}

	/*
	 * initializes a ModelStructureQueryConnector- and a
	 * PerformanceMetricsConnector instance. Both are needed for executing a
	 * WhatIfQuery.
	 */
	private void initializeConnectors(String connectorFamilyName) {
		this.modelStructureQuerConnector = WhatIfUtil.getMsqc(connectorFamilyName);
		this.performanceMetricsConnector = WhatIfUtil.getPmqc(connectorFamilyName);
		for (QueryProgressListener listener : listeners)
			this.performanceMetricsConnector.addQueryProgressListener(listener);
	}

	/*
	 * creates DoFs for executing with the PerformanceMetricsQueryConnector
	 */
	private EntityMapping addDoFToRequest(final EntityMapping requestWithoutDoF) {
		EntityMapping requestWithDoF = EcoreUtil.copy(requestWithoutDoF);
		HashMap<String, List<Integer>> dofIntervals = WhatIfUtil.processWhatIfVariation(query, getInitialValues());

		for (Entry<String, List<Integer>> entry : dofIntervals.entrySet()) {
			DoF dof = domainFactory.createDoF();
			dof.setIdentifier(entry.getKey());
			StringBuilder builder = new StringBuilder("[");

			for (int value : entry.getValue())
				builder.append(value + ", ");
			builder.delete(builder.length() - 2, builder.length());
			builder.append("]");

			dof.setProperties(new Properties());
			dof.getProperties().put("value", builder.toString());

			requestWithDoF.getDoF().add(dof);
		}
		return requestWithDoF;
	}

	/*
	 * get initial values from ModelStructureQueryConnector. Connector gets an
	 * empty EntityMapping to add the DoFs with their initialValues. Those are
	 * put into the HashMap
	 */
	private HashMap<String, Integer> getInitialValues() {
		HashMap<String, Integer> initialValues = new HashMap<String, Integer>();
		EntityMapping mapping = modelStructureQuerConnector.requestDoF(createEmptyEntityMapping());
		for (DoF doF : mapping.getDoF()) {
			initialValues.put(doF.getIdentifier(), Integer.valueOf(doF.getCurrentState()));
		}
		return initialValues;
	}

	/*
	 * creates an EntityMapping with requested metrics. If there are no special
	 * metrics requested, all metrics the connector provieds are used
	 */
	private EntityMapping createRequestWithoutDof() {
		if (query.getMetricReferenceClauses() != null)
			return getEntityMappingForSpecificMetrics();
		else
			return getEntityMappingForAllMetrics();
	}

	/*
	 * creates an EntityMapping for all metrics the ModelStructureQueryConnector
	 * provifdes
	 */
	private EntityMapping getEntityMappingForAllMetrics() {
		EntityMapping mapping = createEmptyEntityMapping();
		mapping = modelStructureQuerConnector.requestEntities(mapping);
		mapping = modelStructureQuerConnector.requestProbes(mapping);
		return mapping;
	}

	/*
	 * created an entityMapping for special metrics user chose in the query.
	 */
	private EntityMapping getEntityMappingForSpecificMetrics() {
		Map<String, Entity> aliasMap = new HashMap<String, Entity>();
		EntityMapping mapping = createEmptyEntityMapping();
		WhatIfUtil.processForClause(query, aliasMap, mapping);
		WhatIfUtil.processSelectClause(query, aliasMap);
		return mapping;
	}

	/*
	 * Method to check if a DoF is supported or not
	 */
	private void validateDoFs() {
		List<String> queryDoFNames = requestQueryDoFNames();
		List<String> existingDoFNames = requestExistingDoFNames();
		for (String identifier : queryDoFNames)
			if (!existingDoFNames.contains(identifier))
				throw new IllegalArgumentException("The DoF " + identifier + " does not exist in SuD, aborting!");
	}

	/*
	 * getting all DoFs from query in order to compare query-DoFs with existing
	 * DoFs
	 */
	private List<String> requestQueryDoFNames() {
		List<String> dofNames = new LinkedList<String>();
		for (WhatIfVariationClause clause : query.getWhatIfVariation().getWhatIfVariationClause())
			dofNames.add(clause.getVariationPoint());
		return dofNames;
	}

	/*
	 * getting all DoFs from connector in order to compare query-DoFs with
	 * existing DoFs
	 */
	private List<String> requestExistingDoFNames() {
		List<String> existingDoFNames = new LinkedList<String>();
		EntityMapping mapping = modelStructureQuerConnector.requestDoF(createEmptyEntityMapping());
		for (DoF dof : mapping.getDoF())
			existingDoFNames.add(dof.getIdentifier());
		return existingDoFNames;
	}

	/*
	 * creates an empty EntityMapping with model location
	 */
	private EntityMapping createEmptyEntityMapping() {
		EntityMapping mapping = mappingFactory.createEntityMapping();
		mapping.setModelLocation(modelLocation);
		return mapping;
	}

}
